/*
 * GPIO /dev and configfs interface
 *
 * Copyright (C) 2006-2007 Atmel Corporation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */
#include <linux/configfs.h>
#include <linux/cdev.h>
#include <linux/fs.h>
#include <linux/interrupt.h>
#include <linux/poll.h>
#include <linux/uaccess.h>
#include <linux/wait.h>

#include <asm/gpio.h>
#include <asm/arch/portmux.h>

#define GPIO_DEV_MAX			8

static struct class *gpio_dev_class;
static dev_t gpio_devt;

struct gpio_item {
	spinlock_t lock;

	int enabled;
	int initialized;
	int port;
	u32 pin_mask;
	u32 oe_mask;

	/* Pin state last time we read it (for blocking reads) */
	u32 pin_state;
	int changed;

	wait_queue_head_t change_wq;
	struct fasync_struct *async_queue;

	int id;
	struct class_device *gpio_dev;
	struct cdev char_dev;
	struct config_item item;
};

struct gpio_attribute {
	struct configfs_attribute attr;
	ssize_t (*show)(struct gpio_item *, char *);
	ssize_t (*store)(struct gpio_item *, const char *, size_t);
};

static irqreturn_t gpio_dev_interrupt(int irq, void *dev_id)
{
	struct gpio_item *gpio = dev_id;
	u32 old_state, new_state;

	old_state = gpio->pin_state;
	new_state = at32_gpio_get_value_multiple(gpio->port, gpio->pin_mask);
	gpio->pin_state = new_state;

	if (new_state != old_state) {
		gpio->changed = 1;
		wake_up_interruptible(&gpio->change_wq);

		if (gpio->async_queue)
			kill_fasync(&gpio->async_queue, SIGIO, POLL_IN);
	}

	return IRQ_HANDLED;
}

static int gpio_dev_open(struct inode *inode, struct file *file)
{
	struct gpio_item *gpio = container_of(inode->i_cdev,
					      struct gpio_item,
					      char_dev);
	unsigned int irq;
	unsigned int i;
	int ret;

	nonseekable_open(inode, file);
	config_item_get(&gpio->item);
	file->private_data = gpio;

	gpio->pin_state = at32_gpio_get_value_multiple(gpio->port,
						       gpio->pin_mask);
	gpio->changed = 1;

	for (i = 0; i < 32; i++) {
		if (gpio->pin_mask & (1 << i)) {
			irq = gpio_to_irq(32 * gpio->port + i);
			ret = request_irq(irq, gpio_dev_interrupt, 0,
					  "gpio-dev", gpio);
			if (ret)
				goto err_irq;
		}
	}

	return 0;

err_irq:
	while (i--) {
		if (gpio->pin_mask & (1 << i)) {
			irq = gpio_to_irq(32 * gpio->port + i);
			free_irq(irq, gpio);
		}
	}

	config_item_put(&gpio->item);

	return ret;
}

static int gpio_dev_fasync(int fd, struct file *file, int mode)
{
	struct gpio_item *gpio = file->private_data;

	return fasync_helper(fd, file, mode, &gpio->async_queue);
}

static int gpio_dev_release(struct inode *inode, struct file *file)
{
	struct gpio_item *gpio = file->private_data;
	unsigned int irq;
	unsigned int i;

	gpio_dev_fasync(-1, file, 0);

	for (i = 0; i < 32; i++) {
		if (gpio->pin_mask & (1 << i)) {
			irq = gpio_to_irq(32 * gpio->port + i);
			free_irq(irq, gpio);
		}
	}

	config_item_put(&gpio->item);

	return 0;
}

static unsigned int gpio_dev_poll(struct file *file, poll_table *wait)
{
	struct gpio_item *gpio = file->private_data;
	unsigned int mask = 0;

	poll_wait(file, &gpio->change_wq, wait);
	if (gpio->changed)
		mask |= POLLIN | POLLRDNORM;

	return mask;
}

static ssize_t gpio_dev_read(struct file *file, char __user *buf,
			     size_t count, loff_t *offset)
{
	struct gpio_item *gpio = file->private_data;
	u32 value;

	spin_lock_irq(&gpio->lock);
	while (!gpio->changed) {
		spin_unlock_irq(&gpio->lock);

		if (file->f_flags & O_NONBLOCK)
			return -EAGAIN;

		if (wait_event_interruptible(gpio->change_wq, gpio->changed))
			return -ERESTARTSYS;

		spin_lock_irq(&gpio->lock);
	}

	gpio->changed = 0;
	value = at32_gpio_get_value_multiple(gpio->port, gpio->pin_mask);

	spin_unlock_irq(&gpio->lock);

	count = min(count, (size_t)4);
	if (copy_to_user(buf, &value, count))
		return -EFAULT;

	return count;
}

static ssize_t gpio_dev_write(struct file *file, const char __user *buf,
			      size_t count, loff_t *offset)
{
	struct gpio_item *gpio = file->private_data;
	u32 value = 0;
	u32 mask = ~0UL;

	count = min(count, (size_t)4);
	if (copy_from_user(&value, buf, count))
		return -EFAULT;

	/* Assuming big endian */
	mask <<= (4 - count) * 8;
	mask &= gpio->pin_mask;

	at32_gpio_set_value_multiple(gpio->port, value, mask);

	return count;
}

static struct file_operations gpio_dev_fops = {
	.owner		= THIS_MODULE,
	.llseek		= no_llseek,
	.open		= gpio_dev_open,
	.release	= gpio_dev_release,
	.fasync		= gpio_dev_fasync,
	.poll		= gpio_dev_poll,
	.read		= gpio_dev_read,
	.write		= gpio_dev_write,
};

static struct gpio_item *to_gpio_item(struct config_item *item)
{
	return item ? container_of(item, struct gpio_item, item) : NULL;
}

static ssize_t gpio_show_gpio_id(struct gpio_item *gpio, char *page)
{
	return sprintf(page, "%d\n", gpio->port);
}

static ssize_t gpio_store_gpio_id(struct gpio_item *gpio,
				  const char *page, size_t count)
{
	unsigned long id;
	char *p = (char *)page;
	ssize_t ret = -EINVAL;

	id = simple_strtoul(p, &p, 0);
	if (!p || (*p && (*p != '\n')))
		return -EINVAL;

	/* Switching PIO is not allowed when live... */
	spin_lock(&gpio->lock);
	if (!gpio->enabled) {
		ret = -ENXIO;
		if (at32_gpio_port_is_valid(id)) {
			gpio->port = id;
			ret = count;
		}
	}
	spin_unlock(&gpio->lock);

	return ret;
}

static ssize_t gpio_show_pin_mask(struct gpio_item *gpio, char *page)
{
	return sprintf(page, "0x%08x\n", gpio->pin_mask);
}

static ssize_t gpio_store_pin_mask(struct gpio_item *gpio,
				   const char *page, size_t count)
{
	u32 new_mask;
	char *p = (char *)page;
	ssize_t ret = -EINVAL;

	new_mask = simple_strtoul(p, &p, 0);
	if (!p || (*p && (*p != '\n')))
		return -EINVAL;

	/* Can't update the pin mask while live. */
	spin_lock(&gpio->lock);
	if (!gpio->enabled) {
		gpio->oe_mask &= new_mask;
		gpio->pin_mask = new_mask;
		ret = count;
	}
	spin_unlock(&gpio->lock);

	return ret;
}

static ssize_t gpio_show_oe_mask(struct gpio_item *gpio, char *page)
{
	return sprintf(page, "0x%08x\n", gpio->oe_mask);
}

static ssize_t gpio_store_oe_mask(struct gpio_item *gpio,
				  const char *page, size_t count)
{
	u32 mask;
	char *p = (char *)page;
	ssize_t ret = -EINVAL;

	mask = simple_strtoul(p, &p, 0);
	if (!p || (*p && (*p != '\n')))
		return -EINVAL;

	spin_lock(&gpio->lock);
	if (!gpio->enabled) {
		gpio->oe_mask = mask & gpio->pin_mask;
		ret = count;
	}
	spin_unlock(&gpio->lock);

	return ret;
}

static ssize_t gpio_show_enabled(struct gpio_item *gpio, char *page)
{
	return sprintf(page, "%d\n", gpio->enabled);
}

static ssize_t gpio_store_enabled(struct gpio_item *gpio,
				  const char *page, size_t count)
{
	char *p = (char *)page;
	int enabled;
	int ret;

	enabled = simple_strtoul(p, &p, 0);
	if (!p || (*p && (*p != '\n')))
		return -EINVAL;

	/* make it a boolean value */
	enabled = !!enabled;

	if (gpio->enabled == enabled)
		/* No change; do nothing. */
		return count;

	BUG_ON(gpio->id >= GPIO_DEV_MAX);

	if (!enabled) {
		class_device_unregister(gpio->gpio_dev);
		cdev_del(&gpio->char_dev);
		at32_deselect_pins(gpio->port, gpio->pin_mask);
		gpio->initialized = 0;
	} else {
		if (gpio->port < 0 || !gpio->pin_mask)
			return -ENODEV;
	}

	/* Disallow any updates to gpio_id or pin_mask */
	spin_lock(&gpio->lock);
	gpio->enabled = enabled;
	spin_unlock(&gpio->lock);

	if (!enabled)
		return count;

	/* Now, try to allocate the pins */
	ret = at32_select_gpio_pins(gpio->port, gpio->pin_mask, gpio->oe_mask);
	if (ret)
		goto err_alloc_pins;

	gpio->initialized = 1;

	cdev_init(&gpio->char_dev, &gpio_dev_fops);
	gpio->char_dev.owner = THIS_MODULE;
	ret = cdev_add(&gpio->char_dev, MKDEV(MAJOR(gpio_devt), gpio->id), 1);
	if (ret < 0)
		goto err_cdev_add;
	gpio->gpio_dev = class_device_create(gpio_dev_class, NULL,
					     MKDEV(MAJOR(gpio_devt), gpio->id),
					     NULL,
					     "gpio%d", gpio->id);
	if (IS_ERR(gpio->gpio_dev)) {
		printk(KERN_ERR "failed to create gpio%d\n", gpio->id);
		ret = PTR_ERR(gpio->gpio_dev);
		goto err_class_dev;
	}

	printk(KERN_INFO "created gpio%d (port%d/0x%08x) as (%d:%d)\n",
	       gpio->id, gpio->port, gpio->pin_mask,
	       MAJOR(gpio->gpio_dev->devt), MINOR(gpio->gpio_dev->devt));

	return 0;

err_class_dev:
	cdev_del(&gpio->char_dev);
err_cdev_add:
	at32_deselect_pins(gpio->port, gpio->pin_mask);
	gpio->initialized = 0;
err_alloc_pins:
	spin_lock(&gpio->lock);
	gpio->enabled = 0;
	spin_unlock(&gpio->lock);

	return ret;
}

static struct gpio_attribute gpio_item_attr_gpio_id = {
	.attr = {
		.ca_owner = THIS_MODULE,
		.ca_name = "gpio_id",
		.ca_mode = S_IRUGO | S_IWUSR,
	},
	.show = gpio_show_gpio_id,
	.store = gpio_store_gpio_id,
};
static struct gpio_attribute gpio_item_attr_pin_mask = {
	.attr = {
		.ca_owner = THIS_MODULE,
		.ca_name = "pin_mask",
		.ca_mode = S_IRUGO | S_IWUSR,
	},
	.show = gpio_show_pin_mask,
	.store = gpio_store_pin_mask,
};
static struct gpio_attribute gpio_item_attr_oe_mask = {
	.attr = {
		.ca_owner = THIS_MODULE,
		.ca_name = "oe_mask",
		.ca_mode = S_IRUGO | S_IWUSR,
	},
	.show = gpio_show_oe_mask,
	.store = gpio_store_oe_mask,
};
static struct gpio_attribute gpio_item_attr_enabled = {
	.attr = {
		.ca_owner = THIS_MODULE,
		.ca_name = "enabled",
		.ca_mode = S_IRUGO | S_IWUSR,
	},
	.show = gpio_show_enabled,
	.store = gpio_store_enabled,
};

static struct configfs_attribute *gpio_item_attrs[] = {
	&gpio_item_attr_gpio_id.attr,
	&gpio_item_attr_pin_mask.attr,
	&gpio_item_attr_oe_mask.attr,
	&gpio_item_attr_enabled.attr,
	NULL,
};

static ssize_t gpio_show_attr(struct config_item *item,
			      struct configfs_attribute *attr,
			      char *page)
{
	struct gpio_item *gpio_item = to_gpio_item(item);
	struct gpio_attribute *gpio_attr
		= container_of(attr, struct gpio_attribute, attr);
	ssize_t ret = 0;

	if (gpio_attr->show)
		ret = gpio_attr->show(gpio_item, page);
	return ret;
}

static ssize_t gpio_store_attr(struct config_item *item,
			       struct configfs_attribute *attr,
			       const char *page, size_t count)
{
	struct gpio_item *gpio_item = to_gpio_item(item);
	struct gpio_attribute *gpio_attr
		= container_of(attr, struct gpio_attribute, attr);
	ssize_t ret = -EINVAL;

	if (gpio_attr->store)
		ret = gpio_attr->store(gpio_item, page, count);
	return ret;
}

static void gpio_release(struct config_item *item)
{
	kfree(to_gpio_item(item));
}

static struct configfs_item_operations gpio_item_ops = {
	.release		= gpio_release,
	.show_attribute		= gpio_show_attr,
	.store_attribute	= gpio_store_attr,
};

static struct config_item_type gpio_item_type = {
	.ct_item_ops	= &gpio_item_ops,
	.ct_attrs	= gpio_item_attrs,
	.ct_owner	= THIS_MODULE,
};

static struct config_item *gpio_make_item(struct config_group *group,
					  const char *name)
{
	static int next_id;
	struct gpio_item *gpio;

	if (next_id >= GPIO_DEV_MAX)
		return NULL;

	gpio = kzalloc(sizeof(struct gpio_item), GFP_KERNEL);
	if (!gpio)
		return NULL;

	gpio->id = next_id++;
	config_item_init_type_name(&gpio->item, name, &gpio_item_type);
	spin_lock_init(&gpio->lock);
	init_waitqueue_head(&gpio->change_wq);

	return &gpio->item;
}

static void gpio_drop_item(struct config_group *group,
			   struct config_item *item)
{
	struct gpio_item *gpio = to_gpio_item(item);

	spin_lock(&gpio->lock);
	if (gpio->enabled) {
		class_device_unregister(gpio->gpio_dev);
		cdev_del(&gpio->char_dev);
	}

	if (gpio->initialized) {
		at32_deselect_pins(gpio->port, gpio->pin_mask);
		gpio->initialized = 0;
		gpio->enabled = 0;
	}
	spin_unlock(&gpio->lock);
}

static struct configfs_group_operations gpio_group_ops = {
	.make_item	= gpio_make_item,
	.drop_item	= gpio_drop_item,
};

static struct config_item_type gpio_group_type = {
	.ct_group_ops	= &gpio_group_ops,
	.ct_owner	= THIS_MODULE,
};

static struct configfs_subsystem gpio_subsys = {
	.su_group = {
		.cg_item = {
			 .ci_namebuf = "gpio",
			 .ci_type = &gpio_group_type,
		 },
	},
};

static int __init gpio_dev_init(void)
{
	int err;

	gpio_dev_class = class_create(THIS_MODULE, "gpio-dev");
	if (IS_ERR(gpio_dev_class)) {
		err = PTR_ERR(gpio_dev_class);
		goto err_class_create;
	}

	err = alloc_chrdev_region(&gpio_devt, 0, GPIO_DEV_MAX, "gpio");
	if (err < 0)
		goto err_alloc_chrdev;

	/* Configfs initialization */
	config_group_init(&gpio_subsys.su_group);
	init_MUTEX(&gpio_subsys.su_sem);
	err = configfs_register_subsystem(&gpio_subsys);
	if (err)
		goto err_register_subsys;

	return 0;

err_register_subsys:
	unregister_chrdev_region(gpio_devt, GPIO_DEV_MAX);
err_alloc_chrdev:
	class_destroy(gpio_dev_class);
err_class_create:
	printk(KERN_WARNING "Failed to initialize gpio /dev interface\n");
	return err;
}
late_initcall(gpio_dev_init);
